import json
import logging
import time

import openai

from actionweaver.action import ActionHandlers
from actionweaver.llms.openai.tokens import TokenUsageTracker


class OpenAIChatCompletionException(Exception):
    pass


class OpenAIChatCompletion:
    def __init__(self, model, token_usage_tracker=None, logger=None):
        self.model = model
        self.action_handlers = ActionHandlers()
        self.logger = logger or logging.getLogger(__name__)
        self.token_usage_tracker = token_usage_tracker or TokenUsageTracker(
            logger=logger
        )

    def _bind_action_handlers(self, action_handlers: ActionHandlers):
        self.action_handlers = action_handlers

    def create(self, messages, *args, scope=None, **kwargs):
        # Restart token usage tracker
        self.token_usage_tracker.clear()

        if scope is None:
            scope = "global"

        self.logger.debug(
            {
                "message": "Creating new chat completion",
                "model": self.model,
                "scope": scope,
                "input_messages": messages,
                "timestamp": time.time(),
            }
        )

        functions = [
            {
                "name": action.name,
                "description": action.description,
                "parameters": action.json_schema(),
            }
            for _, action in self.action_handlers.scope(scope).items()
        ]

        if len(functions) == 0:
            self.logger.debug(
                {
                    "message": "Calling OpenAI api without functions",
                    "input_messages": messages,
                    "timestamp": time.time(),
                }
            )
            api_response = openai.ChatCompletion.create(
                model=self.model, messages=messages, *args, **kwargs
            )
            self.token_usage_tracker.track_usage(api_response.usage)

            self.logger.debug(
                {
                    "message": "Received response from OpenAI api",
                    "response": api_response,
                    "model": self.model,
                    "timestamp": time.time(),
                }
            )
            message = api_response.choices[0]["message"]

            if "content" in message and message["content"]:
                response = message["content"]
                messages += [{"role": "assistant", "content": message["content"]}]
                self.logger.debug(
                    {
                        "message": "Processing message response",
                        "role": message["role"],
                        "content": message["content"],
                        "model": self.model,
                        "timestamp": time.time(),
                    }
                )
            else:
                self.logger.error(
                    {
                        "message": "Unsupported response from OpenAI api",
                        "timestamp": time.time(),
                    },
                    exc_info=True,
                )

                raise OpenAIChatCompletionException(
                    "Unsupported response from OpenAI api"
                )
            return response
        else:
            return self.function_dispatch_loop(
                messages, functions, *args, scope=scope, **kwargs
            )

    def function_dispatch_loop(self, messages, functions, *args, scope=None, **kwargs):
        response = None
        while True:
            self.logger.debug(
                {
                    "message": "Calling OpenAI api with functions",
                    "input_messages": messages,
                    "model": self.model,
                    "scope": scope,
                    "timestamp": time.time(),
                    "functions": [func["name"] for func in functions],
                }
            )

            api_response = openai.ChatCompletion.create(
                model=self.model,
                messages=messages,
                functions=functions,
                *args,
                **kwargs,
            )

            self.token_usage_tracker.track_usage(api_response.usage)

            self.logger.debug(
                {
                    "message": "Received response from OpenAI api",
                    "response": api_response,
                    "timestamp": time.time(),
                }
            )

            message = api_response.choices[0]["message"]

            if "content" in message and message["content"]:
                response = message["content"]
                messages += [{"role": "assistant", "content": message["content"]}]
                self.logger.debug(
                    {
                        "message": "Processing message response",
                        "role": message["role"],
                        "content": message["content"],
                        "model": self.model,
                        "timestamp": time.time(),
                    }
                )
                break
            elif "function_call" in message and message["function_call"]:
                messages += [
                    {
                        "role": "assistant",
                        "content": None,
                        "function_call": message["function_call"].to_dict(),
                    }
                ]
                name = message["function_call"]["name"]
                arguments = message["function_call"]["arguments"]

                self.logger.debug(
                    {
                        "message": "Processing function call",
                        "model": self.model,
                        "function_name": name,
                        "function_arguments": arguments,
                        "timestamp": time.time(),
                    }
                )

                if self.action_handlers.contains(name):
                    try:
                        arguments = json.loads(message["function_call"]["arguments"])
                    except json.decoder.JSONDecodeError as e:
                        self.logger.error(
                            {
                                "message": "Parsing function call arguments from OpenAi response ",
                                "arguments": message["function_call"]["arguments"],
                                "timestamp": time.time(),
                                "model": self.model,
                            },
                            exc_info=True,
                        )
                        raise OpenAIChatCompletionException(e) from e

                    # Invoke action
                    function_response = self.action_handlers[name](**arguments)
                    messages += [
                        {
                            "role": "function",
                            "name": name,
                            "content": str(function_response),
                        }
                    ]

                    self.logger.debug(
                        {
                            "message": "Action invoked and response received",
                            "action_name": name,
                            "action_arguments": arguments,
                            "action_response": function_response,
                            "timestamp": time.time(),
                        }
                    )
                else:
                    unavailable_function_msg = f"{name} is not a valid function name, use one of the following: {', '.join([func['name'] for func in functions])}"
                    messages += [
                        {
                            "role": "user",
                            "content": unavailable_function_msg,
                        }
                    ]
                    self.logger.debug(
                        {
                            "message": "Unavailable action",
                            "action_name": name,
                            "action_arguments": arguments,
                            "action_response": unavailable_function_msg,
                            "timestamp": time.time(),
                        }
                    )
            else:
                raise OpenAIChatCompletionException(
                    "Unsupported response from OpenAI api"
                )
        return response



from __future__ import annotations

import logging
from typing import Any, Dict

from actionweaver.utils import create_pydantic_model_from_func


class ActionException(Exception):
    pass


def action(name, scope="global", logger=None, models=[]):
    """
    Decorator function to create an Action object.

    Args:
    - name (str): Name of the action.
    - scope (str): Scope of the action, default is "global".
    - logger (logging.Logger): Logger instance to log information, default is None.
    - models (list[pydantic.BaseModel]): List of pydantic models to be used in the action.

    Returns:
    - create_action: A function that takes a decorated object and returns an Action object.
    """

    _logger = logger or logging.getLogger(__name__)

    def create_action(decorated_obj):
        _logger.debug({"message": f"Creating action with name: {name}, scope: {scope}"})

        action = Action(
            name=name, scope=scope, decorated_obj=decorated_obj, logger=_logger
        ).build_pydantic_model_cls(models=models)

        return action

    return create_action


class Action:
    def __init__(
        self,
        name,
        scope,
        decorated_obj,
        logger,
    ):
        self.name = name
        self.scope = scope
        self.logger = logger

        if decorated_obj.__doc__ is None:
            raise ActionException(
                f"Decorated method under action {name} must have a docstring for description."
            )

        self.description = decorated_obj.__doc__
        self.pydantic_cls = None

        self.decorated_method = decorated_obj

    def build_pydantic_model_cls(self, models=None):
        if models is None:
            models = []

        self.pydantic_cls = create_pydantic_model_from_func(
            self.decorated_method, self.decorated_method.__name__.title(), models=models
        )
        return self

    def json_schema(self):
        return self.pydantic_cls.model_json_schema()

    def bind(self, instance) -> InstanceAction:
        return InstanceAction(self, instance)

    def __get__(self, instance, owner) -> InstanceAction:
        """
        Bind a action with an instance.

        Note:
            The `__get__` method is a descriptor method that is called when the action is accessed from an instance.
            It returns an instance-specific action method that is bound to the given instance.
        """
        return InstanceAction(self, instance)


class InstanceAction:
    def __init__(self, action, instance):
        self.action = action
        self.__module__ = action.decorated_method.__module__
        self.__name__ = action.decorated_method.__name__
        self.__qualname__ = action.decorated_method.__qualname__
        self.__annotations__ = action.decorated_method.__annotations__
        self.__doc__ = action.decorated_method.__doc__
        self.instance = instance

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        self.action.logger.debug(
            {
                "message": f"[Action {self.action.name}, method {self.__name__}] Calling action: {self.action.name} with args: {args}"
            }
        )
        response = self.action.decorated_method(self.instance, *args, **kwargs)
        self.action.logger.debug(
            {
                "message": f"[Action {self.action.name}, method {self.__name__}] Received response: {response}"
            }
        )
        return response


class ActionHandlers:
    def __init__(self, *args, **kwargs):
        self.name_to_action: Dict[str, Action] = {}

    def contains(self, name) -> bool:
        return name in self.name_to_action

    def __len__(self) -> int:
        return len(self.name_to_action)

    def scope(self, scope):
        return {
            name: action
            for name, action in self.name_to_action.items()
            if action.scope == scope
        }

    def bind(self, instance) -> InstanceActionHandlers:
        return InstanceActionHandlers(instance, self)

    @classmethod
    def merge(cls, *handlers) -> ActionHandlers:
        merged = cls()
        for handler in handlers:
            merged.name_to_action.update(handler.name_to_action)
        return merged


class InstanceActionHandlers:
    def __init__(self, instance, action_handlers: ActionHandlers, *args, **kwargs):
        self.action_handlers = action_handlers
        self.instance = instance

    def __getitem__(self, key) -> InstanceAction:
        val = self.action_handlers.name_to_action[key]
        return val.bind(self.instance)

    def __len__(self) -> int:
        return len(self.action_handlers)

    def scope(self, scope):
        return self.action_handlers.scope(scope)

    def contains(self, name) -> bool:
        return self.action_handlers.contains(name)



from __future__ import annotations

import copy

from actionweaver.action import Action, ActionHandlers
from actionweaver.llms.openai.chat import OpenAIChatCompletion


class ActionHandlerMixin:
    _action_handlers = ActionHandlers()

    def __post_init__(self):
        # bind action handlers to self
        self.instance_action_handlers = self._action_handlers.bind(self)

        for _, attr_value in tuple(self.__dict__.items()):
            if isinstance(attr_value, OpenAIChatCompletion):
                # bind instance action handlers to llm
                attr_value._bind_action_handlers(self.instance_action_handlers)

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)

        def init_decorator(original_init):
            def new_init(self, *args, **kwargs):
                original_init(self, *args, **kwargs)
                if type(self) == cls:
                    self.__post_init__()

            return new_init

        # Add __post_init__ method to the subclass.
        cls.__init__ = init_decorator(cls.__init__)

        # create action handlers from base class.
        cls._action_handlers = ActionHandlers.merge(
            *[
                base_cls._action_handlers
                for base_cls in cls.__bases__
                if hasattr(base_cls, "_action_handlers")
            ]
        )

        for _, attr_value in tuple(cls.__dict__.items()):
            if isinstance(attr_value, Action):
                cls._action_handlers.name_to_action[attr_value.name] = attr_value
